from contextlib import contextmanager
from functools import wraps
from string import Template
from typing import Callable, Iterable, Iterator, Type

from pydoctrace.callfilter import Preset, call_filter_factory
from pydoctrace.callfilter.presets import EXCLUDE_STDLIB_PRESET, EXCLUDE_TESTS_PRESET
from pydoctrace.exporters import Context, Exporter
from pydoctrace.exporters.plantuml.component import PlantUMLComponentExporter
from pydoctrace.exporters.plantuml.sequence import PlantUMLSequenceExporter
from pydoctrace.tracer import ExecutionTracer

# default filters used to remove calls from the execution tracing
DEFAULT_FILTERS = EXCLUDE_STDLIB_PRESET, EXCLUDE_TESTS_PRESET


@contextmanager
def tracing_context_factory(context: Context) -> Iterator[ExecutionTracer]:
    with context.exporter_class.export_manager_factory(context.export_file_path) as exporter:
        # initializes the diagram file
        exporter.on_header(context.start_module, context.start_function_name)

        tracer = ExecutionTracer(exporter, context.call_filter)
        try:
            yield tracer
        finally:
            # finalizes the sequence diagram file
            exporter.on_footer()


def context_factory(
    function_to_trace: Callable,
    exporter_class: Type[Exporter],
    export_file_path_tpl: str,
    filter_presets: Iterable[Preset] = None,
) -> Context:
    function_module = getattr(function_to_trace, '__module__', '__root__')
    function_name = getattr(function_to_trace, '__name__', '__main__')
    export_file_path = Template(export_file_path_tpl).safe_substitute(
        function_name=function_name, function_module=function_module
    )
    filter_presets = DEFAULT_FILTERS if filter_presets is None else filter_presets

    return Context(
        exporter_class, export_file_path, function_module, function_name, call_filter_factory(filter_presets)
    )


def trace_to_sequence_puml(
    function_to_decorate: Callable = None,
    /,
    *,
    export_file_path_tpl: str = '${function_name}-sequence.puml',
    filter_presets: Iterable[Preset] = None,
):
    """
    Decorates a function in order to trace its execution as a sequence diagram.
    - filter_presets: enable to remove specific calls from the execution tracing. Use provided presets or design yours.
      By defaults (if None), filters out calls to the tests related modules and standard library modules.
      Set to an empty iterable to disable call filtering.

    - export_file_path_tpl: customizes the file path where the output will be written to.
      It can include placeholders like '${function_module}', '${function_name}', ${datetime_millis}'.
    """

    def sequence_puml_decorator(function_to_trace: Callable):
        @wraps(function_to_trace)
        def traceable_func(*args, **kwargs):
            # initializes the tracing context
            context = context_factory(function_to_trace, PlantUMLSequenceExporter, export_file_path_tpl, filter_presets)

            # runs the decorated function in a tracing context
            with tracing_context_factory(context) as execution_tracer:
                return execution_tracer.runfunc(function_to_trace, *args, **kwargs)

        return traceable_func

    if function_to_decorate:
        return sequence_puml_decorator(function_to_decorate)
    else:
        return sequence_puml_decorator


def trace_to_component_puml(
    function_to_decorate: Callable = None,
    /,
    *,
    export_file_path_tpl: str = '${function_name}-component.puml',
    filter_presets: Iterable[Preset] = None,
):
    """
    Decorates a function in order to trace its execution as a component diagram.
    - filter_presets: enable to remove specific calls from the execution tracing. Use provided presets or design yours.
      By defaults (if None), filters out calls to the tests related modules and standard library modules.
      Set to an empty iterable to disable call filtering.

    - export_file_path_tpl: customizes the file path where the output will be written to.
      It can include placeholders like '${function_module}', '${function_name}', ${datetime_millis}'.
    """

    def component_puml_decorator(function_to_trace: Callable):
        @wraps(function_to_trace)
        def traceable_func(*args, **kwargs):
            # initializes the tracing context
            context = context_factory(
                function_to_trace, PlantUMLComponentExporter, export_file_path_tpl, filter_presets
            )

            # runs the decorated function in a tracing context
            with tracing_context_factory(context) as execution_tracer:
                return execution_tracer.runfunc(function_to_trace, *args, **kwargs)

        return traceable_func

    if function_to_decorate:
        return component_puml_decorator(function_to_decorate)
    else:
        return component_puml_decorator



"""
This module provides formatting features to be used when producing the diagram code.
Some characters that can appear in function or module names can be interpreted by
the diagram syntax as formatting characters and thus should be replaced by unicode
entities or escaped.

The module proposes low-level functions adapted to be used in a Python string.Formatter
(see https://docs.python.org/3/library/string.html#string.Formatter) and follow
the signature of the Formatter.format_field(self, value: Any, format_spec: str) method
(see https://docs.python.org/3/library/string.html#string.Formatter.format_field).

These formatting functions can be then used as reusable components to build custom
formatters that suit specific needs only (depending on the diagram syntax).

In the string template to format, you can define and use specific format_spec flag
to activate a formatting feature of the formatter. For example, if you know that
some '__' dunder sdtrings need to be escaped, you can express it in the template:
CALL_TPL = r'{calling_function.name:dunder} -> {called_function.name:dunder}'

Once the formatting component has handled the dunder thing, the 'dunder' format_spec
must be replaced by '' when passed to the parent Formatter.format_field() method
because it would expect something supported by the Python formatting mini-language
(see https://docs.python.org/3/library/string.html#formatspec). Otherwise, an error
will be raised.
"""

from functools import reduce
from re import Pattern
from re import compile as re_compile
from string import Formatter
from typing import Any, Callable, Tuple

AROBASE_REPLACE_PATTERN: Pattern = re_compile('@')
AROBASE_IN_UNICODE: str = '<U+0040>'

DUNDER_REPLACE_PATTERN: Pattern = re_compile('__')


def replace_arobase_by_unicode(value: Any, format_spec: str) -> Tuple[Any, str]:
    """
    Replaces the '@' character by its unicode equivalent
    """
    if value is not None and isinstance(value, str):
        value = AROBASE_REPLACE_PATTERN.sub(AROBASE_IN_UNICODE, value)

    return value, format_spec


def escape_dunder_with_tilde(value: Any, format_spec: str) -> Tuple[Any, str]:
    if format_spec == 'dunder':
        # escapes each text part of the given str value (splitted on '.') with '~' when format_spec is 'dunder'
        # to avoid PlantUML interpreting dunder symbols as striking formatting
        if isinstance(value, str):
            value = DUNDER_REPLACE_PATTERN.sub('~__', value)

        # dunder formatting is done, cancel it for parent formatting
        format_spec = ''

    return value, format_spec


def formatter_factory(formatter_class_name: str, *formatters: Callable[[Any, str], Tuple[Any, str]]) -> Formatter:
    """
    Creates a custom string.Formatter with the given formatters as components.
    The formatters will be applied in the given order.
    """

    def custom_format_field(self: Formatter, value: Any, format_spec: str):
        """
        Applies all the formatters on the given value and format_spec, which are updated by each formatter.
        """
        updated_value, updated_format_spec = reduce(
            lambda value_and_format_spec, formatter: formatter(value_and_format_spec[0], value_and_format_spec[1]),
            formatters,
            (value, format_spec),
        )

        return super(self.__class__, self).format_field(updated_value, updated_format_spec)

    custom_formatter_class = type(formatter_class_name, (Formatter,), {'format_field': custom_format_field})

    return custom_formatter_class()



"""
Module responsible for the execution of a function in a tracing context.

The tracing is based on the sys.settrace hook system, which takes:
- a global tracing function intercepting the calls to block of codes
- a local tracing function, returned by the global tracing function, handling:
  - the line executions (ignored by this tool)
  - the exit of the function (when returning a value or raising an error)
- an exception tracing function, returned by the local tracing function, handling either the error propagation
  or its handling by a try-except block
"""

from collections import deque
from pathlib import Path
from sys import gettrace, settrace
from typing import Any, Callable, List, NamedTuple

from pydoctrace.callfilter import CallFilter
from pydoctrace.domain.execution import CallEnd, Error
from pydoctrace.exporters import Exporter


class TracedError(NamedTuple):
    """
    An error being traced by the tracer, holding metadata:
    - error: Error: information for display purposes in the diagram
    - line_index: int: the line at which the error entered the execution frame
    - exception: BaseException: the corresponding Python exception being traced
    """

    error: Error
    line_index: int


def module_name_from_filepath(script_filepath: str) -> str:
    """Return a plausible module name for the script_filepath."""
    # base = os.path.basename(script_filepath)
    # filename, _ = os.path.splitext(base)
    # return filename
    return None if script_filepath is None else Path(script_filepath).stem


class ExecutionTracer:
    """
    Traces the execution of a callable object and pushes events to the given exporter.

    The implementation of the tracing functions are based on the documentation of:
    - sys.settrace: https://docs.python.org/3/library/sys.html#sys.settrace
    - execution frames: https://docs.python.org/3/reference/datamodel.html#frame-objects

    Useful calls when implementing or debugging features (the contents are added as PlantUML comments in the exported diagram):
    self.exporter.on_raw_content(f"\n' {event} {frame=} {arg=}\n")
    self.exporter.on_raw_content(f"' {frame.f_back=}\n")
    """

    def __init__(self, exporter: Exporter, call_filter: CallFilter):
        self.exporter = exporter
        self.call_filter = call_filter
        self.callers_stack: List[CallEnd] = deque()
        self.error_to_handle_with_line: TracedError = None

    def runfunc(self, func: Callable, *args, **kwargs) -> Any:
        """
        Runs the given function with the given positional and keyword arguments and traces the calls sequence.
        """

        # ensures that a callable object has been passed
        if func is None or not callable(func):
            raise ValueError('A function or a callable object must be passed to trace its execution')

        # declares the tracing callbacks to trace calls, performs and traces the call, then removes the tracer
        tracing_function = gettrace()
        settrace(self.globaltrace)
        try:
            return func(*args, **kwargs)
        finally:
            settrace(tracing_function)

    def on_return_or_exit(self, called_end: CallEnd, arg: Any):
        # the calls stack is empty -> end of the tracing
        if len(self.callers_stack) == 0:
            self.exporter.on_tracing_end(called_end, arg)
        # normal return call
        else:
            self.exporter.on_return(called=called_end, caller=self.callers_stack[-1], arg=arg)

    def error_from_exception(self, exception: BaseException) -> Error:
        error_args = getattr(exception, 'args', None)
        error_message = repr(exception) if error_args is None or len(error_args) == 0 else error_args[0]

        return Error(exception.__class__.__name__, error_message)

    def globaltrace(self, frame, event: str, arg: Any):
        """
        Handler for call events.

        Returns:
        - a custom tracing function to trace the execution of the block
        - None if the execution block should be ignored
        """

        if event == 'call':
            # determines whether the call should be traced or not
            fq_module_text: str = frame.f_globals.get('__name__', None)
            if fq_module_text is None:
                fq_module_text = module_name_from_filepath(frame.f_globals.get('__file__', None))
            fq_module_parts = tuple(fq_module_text.split('.'))
            function_name = frame.f_code.co_name
            if not self.call_filter.should_trace_call(fq_module_parts, function_name, len(self.callers_stack)):
                return None

            # constructs the call
            line_index = frame.f_lineno
            call = CallEnd(fq_module_text, fq_module_parts, function_name, line_index)

            # starts the tracing or handle an intermediary call
            if len(self.callers_stack) == 0:
                self.exporter.on_tracing_start(call)
            else:
                self.exporter.on_start_call(self.callers_stack[-1]._replace(line_index=frame.f_back.f_lineno), call)

            # unflags any error remaining from localtrace without return event
            self.error_to_handle_with_line = None

            # adds the call to the calls stack and returns the call tracing function to detect 'return' or 'exception' events
            self.callers_stack.append(call)

            return self.localtrace

    def localtrace(self, frame, event: str, arg: Any):
        """
        Handler for events happening within a call:
        - 'return' event: ends the current call, removes it from the callers stack

        When an 'exception' occurs:
        - localtrace:exception and exceptiontrace:return have the same line number (localtrace:return is not called) -> propagate the error to the caller
        - localtrace:exception and localtrace:return have the same line number (exceptiontrace:return is not called) -> propagate the error to the caller
        - localtrace:exception and exceptiontrace:return have different line numbers (localtrace:return is not called) -> classic return to the caller

        In all cases:
        - localtrace:exception should flag the error and the line at which it occured (in the tracer)
        - exceptiontrace:exception overrides the previous error (happens when an error is caught and another one is raised in an except block)
        - localtrace:return (when an error is flagged in the tracer) or exceptiontrace:return (there should always be an error in the tracer):
          - if their line numbers are the same -> it is an error propagation
          - if their line numbers are different -> the error has been handled, it is a classic return event
        """

        if event == 'exception':
            # creates the error and flags it so that it can be handled either by the localtrace or exceptiontrace
            error = self.error_from_exception(arg[1])
            self.error_to_handle_with_line = TracedError(error, frame.f_lineno)

            # uses the exception tracing
            return self.exceptiontrace

        elif event == 'return':
            # classic return when the error has been handled internally (in an except block)
            if self.error_to_handle_with_line is None:
                # end of the block code: removes the last caller from the stack
                called_end = self.callers_stack.pop()._replace(line_index=frame.f_lineno)
                self.on_return_or_exit(called_end, arg)
            # propagates the error to the caller
            else:
                error, _ = self.error_to_handle_with_line
                error_called = self.callers_stack.pop()._replace(line_index=frame.f_lineno)
                error_caller = self.callers_stack[-1]._replace(line_index=frame.f_back.f_lineno)
                self.exporter.on_error_propagation(error_called, error_caller, error)

                # unflags the error
                self.error_to_handle_with_line = None

        return self.localtrace

    def exceptiontrace(self, frame, event: str, arg: Any):
        # a new exception enters the frame (re-raised, replaced or wrapped)
        if event == 'exception':
            error = self.error_from_exception(arg[1])
            self.error_to_handle_with_line = TracedError(error, frame.f_lineno)

        elif event == 'return':
            # the error has been internally handled, a classic return occurs
            if self.error_to_handle_with_line is None:
                called_end = self.callers_stack.pop()._replace(line_index=frame.f_lineno)
                self.on_return_or_exit(called_end, arg)
            # handles the flagged error
            else:
                error, line_number_called = self.error_to_handle_with_line
                line_number = frame.f_lineno

                # error propagation
                if line_number == line_number_called:
                    error_called = self.callers_stack.pop()._replace(line_index=line_number)
                    # exits the tracing if there is no caller anymore
                    if len(self.callers_stack) == 0:
                        self.exporter.on_unhandled_error_end(error_called, error)
                    else:
                        error_caller = self.callers_stack[-1]._replace(line_index=frame.f_back.f_lineno)
                        self.exporter.on_error_propagation(error_called, error_caller, error)

                # error was handled, normal return
                else:
                    called_end = self.callers_stack.pop()._replace(line_index=frame.f_lineno)
                    self.on_return_or_exit(called_end, arg)

                # unflags the error
                self.error_to_handle_with_line = None

        return self.exceptiontrace



"""
Module dedicated to the export of sequence diagrams in the PlanUML syntax.

Bibliography:
- https://plantuml.com/en/sequence-diagram: syntax of sequence diagrams
- https://plantuml.com/en/guide, https://plantuml.com/fr/guide: the PlanUML syntax books. Pages or sections of interest:
  - 24.6 Colors: color names
- https://plantuml-documentation.readthedocs.io/en/latest/formatting/all-skin-params.html: "All skin parameters available in PlantUML in alphabetical order."
"""

from string import Formatter
from typing import Any

from pydoctrace.domain.execution import CallEnd, Error
from pydoctrace.exporters import Exporter
from pydoctrace.exporters.formatters import escape_dunder_with_tilde, formatter_factory, replace_arobase_by_unicode
from pydoctrace.exporters.plantuml import FOOTER_TPL

PLANTUML_SEQUENCE_FORMATTER: Formatter = formatter_factory(
    'PlantUMLSequenceFormatter', replace_arobase_by_unicode, escape_dunder_with_tilde
)

HEADER_TPL = r"""@startuml {diagram_name}
skinparam BoxPadding 10
skinparam ParticipantPadding 5
skinparam NoteBackgroundColor Cornsilk
skinparam NoteBorderColor Sienna
hide footbox
"""

TRACING_START_TPL = r"""
[o-> "{called.fq_module_text:dunder}\n{called.function_name:dunder}"
note right: line {called.line_index}
"""

CALL_START_TPL = r"""
"{caller.fq_module_text:dunder}\n{caller.function_name:dunder}" -> "{called.fq_module_text:dunder}\n{called.function_name:dunder}" ++
note left: line {caller.line_index}
note right: line {called.line_index}
"""

CALL_END_TPL = r"""
return {arg:dunder}
note right: line {called.line_index}
|||
"""

ERROR_PROPAGATION_TPL = r"""
"{error_caller.fq_module_text:dunder}\n{error_caller.function_name:dunder}" o<--x "{error_called.fq_module_text:dunder}\n{error_called.function_name:dunder}": ""{error.class_name}""\n{error.message}
deactivate "{error_called.fq_module_text:dunder}\n{error_called.function_name:dunder}"
note right: line {error_called.line_index}
note left: line {error_caller.line_index}
"""

TRACING_END_TPL = r"""
[<-- "{called.fq_module_text:dunder}\n{called.function_name:dunder}": {arg:dunder}
note right: line {called.line_index}
"""

UNHANDLED_ERROR_END_TPL = r"""
[<-->x "{called.fq_module_text:dunder}\n{called.function_name:dunder}": ""{error.class_name}""\n{error.message}
note right: line {called.line_index}
"""


class PlantUMLSequenceExporter(Exporter):
    """
    Exports the sequence diagram in the PlantUML format.

    The nature of the sequence diagram and the corresponding PlantUML syntax
    allows to produce the diagram contents as the execution goes, in a
    streaming fashion.

    For efficiency reasons, instances of the execution domain model are used
    directly without converting them into diagram model instances.
    """

    fmt: Formatter = PLANTUML_SEQUENCE_FORMATTER

    def on_header(self, start_module: str, start_func_name: str):
        diagram_name = f'{start_module}.{start_func_name}-sequence'
        self.io_sink.write(self.fmt.format(HEADER_TPL, diagram_name=diagram_name))

    def on_tracing_start(self, called: CallEnd):
        self.io_sink.write(self.fmt.format(TRACING_START_TPL, called=called))

    def on_start_call(self, caller: CallEnd, called: CallEnd):
        self.io_sink.write(self.fmt.format(CALL_START_TPL, caller=caller, called=called))

    def format_arg_value(self, arg: Any) -> Any:
        if arg is None:
            return ''
        return arg

    def on_error_propagation(self, error_called: CallEnd, error_caller: CallEnd, error: Error):
        self.io_sink.write(
            self.fmt.format(ERROR_PROPAGATION_TPL, error_called=error_called, error_caller=error_caller, error=error)
        )

    def on_return(self, *, called: CallEnd, arg: Any, **kwargs):
        self.io_sink.write(self.fmt.format(CALL_END_TPL, called=called, arg=self.format_arg_value(arg)))

    def on_tracing_end(self, called: CallEnd, arg: Any):
        self.io_sink.write(self.fmt.format(TRACING_END_TPL, called=called, arg=self.format_arg_value(arg)))

    def on_unhandled_error_end(self, called: CallEnd, error: Error):
        self.io_sink.write(self.fmt.format(UNHANDLED_ERROR_END_TPL, called=called, error=error))

    def on_footer(self):
        self.io_sink.write(FOOTER_TPL)



"""
Domain models involved while tracing the code execution.

NamedTuples are used for immutability sake and for their light weight.
"""

from typing import NamedTuple, Tuple


class CallEnd(NamedTuple):
    """
    When a call is made, it is between a "calling end" (at a given line in a given module where a is function being executed)
    and a "called end" (a function whose body starts at a given line in a given module).
    """

    fq_module_text: str
    fq_module_tuple: Tuple[str]
    function_name: str
    line_index: int


class Error(NamedTuple):
    """
    Represents an error or an exception that is raised during the code execution.
    """

    class_name: str
    message: str



from typing import Iterable, Tuple

from pydoctrace.callfilter.presets import EXCLUDE_STDLIB_PRESET, Preset


class CallFilter:
    """
    Offers the possibility to exclude a call from the tracing process to make the diagram more concise.

    The method should_trace_call tells whether the call should be traced. The filter internally uses presets
    to help the user define exclusion rules. A preset gives the possibility to express large-grain exclusion rules
    with fine-grain inclusion rules.
    """

    def __init__(self, presets: Iterable[Preset]):
        self.presets: Tuple[Preset] = tuple(presets)

    def should_trace_call(self, module_parts: Tuple[str], function_name: str, call_depth: int) -> bool:
        """
        Iterates over the presets until one excludes the call:
        - once a preset excludes the call, the filter stops iterating over the presets
        - if no exclusion preset is found, the call should be traced
        """
        call_excluded = any(
            (
                # the call is excluded by the exclusion rule
                preset.exclude_call(module_parts, function_name, call_depth)
                and (
                    # no exception rule is set to include the call
                    preset.include_call is None
                    or
                    # the exception rule does not include the call
                    not preset.include_call(module_parts, function_name, call_depth)
                )
                for preset in self.presets
            )
        )

        return not call_excluded


# The following filters are defined for testing purposes; the tracing decorators involve presets instead of filters.


class TraceAll(CallFilter):
    """
    A light-weight implementation which filters nothing out (traces everything).
    """

    def __init__(self):
        """
        Initializes the parent class with an emtpy tuple of filtering presets
        """
        super().__init__(())

    def should_trace_call(self, module_parts: Tuple[str], function_name: str, call_depth: int) -> bool:
        return True


# singleton instance
TRACE_ALL_FILTER = TraceAll()


class TraceNone(CallFilter):
    """
    A light-weight implementation which filters everything out (traces nothing).
    """

    def __init__(self):
        """
        Initializes the parent class with an emtpy tuple of filtering presets
        """
        super().__init__(())

    def should_trace_call(self, module_parts: Tuple[str], function_name: str, call_depth: int) -> bool:
        return False


# singleton instance
TRACE_NONE_FILTER = TraceNone()


def call_filter_factory(filter_presets: Iterable[Preset]) -> CallFilter:
    """
    Utility function returning an instance of CallFilter ready to be used by the tracer,
    from the given list of presets.
    """
    if filter_presets is None or len(filter_presets) == 0:
        return TRACE_ALL_FILTER
    else:
        return CallFilter(filter_presets)


FILTER_OUT_STDLIB = call_filter_factory((EXCLUDE_STDLIB_PRESET,))



from sys import builtin_module_names, version_info
from typing import Callable, NamedTuple, Optional, Tuple

# the type alias for an exclusion or an inclusion filter. It is a callable taking the following parameters
# and returning a boolean value:
# - Tuple[str]: represents the module fully-qualified name holding the function
# - str: the function name
# - int: the call depth, incremented when a function calls another function, decremented when a function call returns
Filter = Callable[[Tuple[str], str, int], bool]


class Preset(NamedTuple):
    """
    Represents a set of two rules, each one being a function:
    - exclude_call should return True to exclude the call from the tracing process, False to trace it
    - include_call is optional and called only if exclude_call returned True, and is expected to return
      True to force the tracing anyway; thus offering some inclusion exceptions to the exclusion rule

    The typical use-case is to exclude the call to all the functions of a given module in exclude_call,
    then to allow the call to some functions in include_call.

    Note: exclude_call must not be None, include_call can be None if the preset offers no exception rule.
    """

    exclude_call: Filter
    include_call: Optional[Filter] = None


# excludes calls to functions that are in the built-in modules
EXCLUDE_BUILTINS_PRESET = Preset(
    exclude_call=lambda module_parts, *args: len(module_parts) > 0 and module_parts[0] in builtin_module_names
)

# excludes calls to functions that are in the standard library modules (builtins and more, like datetime)
# note: it is based on sys.stdlib_module_names, which appeared in Python 3.10
if version_info.major >= 3 and version_info.minor >= 10:
    from sys import stdlib_module_names
else:
    # list of module names of the Python 3.10 standard library. Some native modules of the previous versions of the
    # standard library may be missing
    stdlib_module_names = (
        '__future__', '_abc', '_aix_support', '_ast', '_asyncio', '_bisect', '_blake2', '_bootsubprocess', '_bz2',
        '_codecs', '_codecs_cn', '_codecs_hk', '_codecs_iso2022', '_codecs_jp', '_codecs_kr', '_codecs_tw',
        '_collections', '_collections_abc', '_compat_pickle', '_compression', '_contextvars', '_crypt', '_csv',
        '_ctypes', '_curses', '_curses_panel', '_datetime', '_dbm', '_decimal', '_elementtree', '_frozen_importlib',
        '_frozen_importlib_external', '_functools', '_gdbm', '_hashlib', '_heapq', '_imp', '_io', '_json', '_locale',
        '_lsprof', '_lzma', '_markupbase', '_md5', '_msi', '_multibytecodec', '_multiprocessing', '_opcode',
        '_operator', '_osx_support', '_overlapped', '_pickle', '_posixshmem', '_posixsubprocess', '_py_abc',
        '_pydecimal', '_pyio', '_queue', '_random', '_scproxy', '_sha1', '_sha256', '_sha3', '_sha512', '_signal',
        '_sitebuiltins', '_socket', '_sqlite3', '_sre', '_ssl', '_stat', '_statistics', '_string', '_strptime',
        '_struct', '_symtable', '_thread', '_threading_local', '_tkinter', '_tokenize', '_tracemalloc', '_typing',
        '_uuid', '_warnings', '_weakref', '_weakrefset', '_winapi', '_zoneinfo', 'abc', 'aifc', 'antigravity',
        'argparse', 'array', 'ast', 'asynchat', 'asyncio', 'asyncore', 'atexit', 'audioop', 'base64', 'bdb', 'binascii',
        'bisect', 'builtins', 'bz2', 'cProfile', 'calendar', 'cgi', 'cgitb', 'chunk', 'cmath', 'cmd', 'code', 'codecs',
        'codeop', 'collections', 'colorsys', 'compileall', 'concurrent', 'configparser', 'contextlib', 'contextvars',
        'copy', 'copyreg', 'crypt', 'csv', 'ctypes', 'curses', 'dataclasses', 'datetime', 'dbm', 'decimal', 'difflib',
        'dis', 'distutils', 'doctest', 'email', 'encodings', 'ensurepip', 'enum', 'errno', 'faulthandler', 'fcntl',
        'filecmp', 'fileinput', 'fnmatch', 'fractions', 'ftplib', 'functools', 'gc', 'genericpath', 'getopt', 'getpass',
        'gettext', 'glob', 'graphlib', 'grp', 'gzip', 'hashlib', 'heapq', 'hmac', 'html', 'http', 'idlelib', 'imaplib',
        'imghdr', 'imp', 'importlib', 'inspect', 'io', 'ipaddress', 'itertools', 'json', 'keyword', 'lib2to3',
        'linecache', 'locale', 'logging', 'lzma', 'mailbox', 'mailcap', 'marshal', 'math', 'mimetypes', 'mmap',
        'modulefinder', 'msilib', 'msvcrt', 'multiprocessing', 'netrc', 'nis', 'nntplib', 'nt', 'ntpath', 'nturl2path',
        'numbers', 'opcode', 'operator', 'optparse', 'os', 'ossaudiodev', 'pathlib', 'pdb', 'pickle', 'pickletools',
        'pipes', 'pkgutil', 'platform', 'plistlib', 'poplib', 'posix', 'posixpath', 'pprint', 'profile', 'pstats',
        'pty', 'pwd', 'py_compile', 'pyclbr', 'pydoc', 'pydoc_data', 'pyexpat', 'queue', 'quopri', 'random', 're',
        'readline', 'reprlib', 'resource', 'rlcompleter', 'runpy', 'sched', 'secrets', 'select', 'selectors', 'shelve',
        'shlex', 'shutil', 'signal', 'site', 'smtpd', 'smtplib', 'sndhdr', 'socket', 'socketserver', 'spwd', 'sqlite3',
        'sre_compile', 'sre_constants', 'sre_parse', 'ssl', 'stat', 'statistics', 'string', 'stringprep', 'struct',
        'subprocess', 'sunau', 'symtable', 'sys', 'sysconfig', 'syslog', 'tabnanny', 'tarfile', 'telnetlib', 'tempfile',
        'termios', 'textwrap', 'this', 'threading', 'time', 'timeit', 'tkinter', 'token', 'tokenize', 'tomllib',
        'trace', 'traceback', 'tracemalloc', 'tty', 'turtle', 'turtledemo', 'types', 'typing', 'unicodedata',
        'unittest', 'urllib', 'uu', 'uuid', 'venv', 'warnings', 'wave', 'weakref', 'webbrowser', 'winreg', 'winsound',
        'wsgiref', 'xdrlib', 'xml', 'xmlrpc', 'zipapp', 'zipfile', 'zipimport', 'zlib', 'zoneinfo',
    )  # fmt: skip

EXCLUDE_STDLIB_PRESET = Preset(
    exclude_call=lambda module_parts, *args: len(module_parts) > 0 and module_parts[0] in stdlib_module_names
)

EXCLUDE_TESTS_PRESET = Preset(
    exclude_call=lambda module_parts, *args: len(module_parts) > 0
    and module_parts[0] in ('tests', '_pytest', 'pytest', 'unittest', 'doctest')
)

TRACE_ALL_PRESET = Preset(exclude_call=lambda *args: False)


def _depth_preset_factory(depth_threshold: int) -> Preset:
    """
    Low-level factory function producing a preset that excludes calls from tracing
    when the call depth is equal or above the given threshold of the calls stack.
    """
    if (depth_threshold is None) or not isinstance(depth_threshold, int):
        raise TypeError('depth threshold must be an integer')
    if depth_threshold < 0:
        raise ValueError(f"depth threshold must be a positive integer, got '{depth_threshold}'")

    return Preset(exclude_call=lambda module_parts, function_name, call_depth: call_depth > depth_threshold)


# exposes the low-level preset factory as an upper-case callable, for the sake of homogeneity with other default presets
EXCLUDE_CALL_DEPTH_PRESET_FACTORY = _depth_preset_factory

# excludes calls whose depth in the calls stack is below (greater than) 5
EXCLUDE_DEPTH_BELOW_5_PRESET = EXCLUDE_CALL_DEPTH_PRESET_FACTORY(5)



from contextlib import contextmanager
from datetime import datetime
from io import TextIOBase
from pathlib import Path
from string import Template
from typing import Any, Iterator, NamedTuple, Type

from pydoctrace.callfilter import CallFilter
from pydoctrace.domain.execution import CallEnd, Error


class Exporter:
    """
    Base class for sequence diagram exporters.
    Extend it to to support various export formats.
    """

    def __init__(self, io_sink: TextIOBase):
        self.io_sink = io_sink

    def on_header(self, start_module: str, start_func_name: str):
        """
        Writes the header of the sequence diagram file.
        This method is supposed to be called only once during an export process.
        """
        raise NotImplementedError()

    def on_raw_content(self, raw_content: str):
        """
        Writes the given raw content to the sink.
        This method is not meant to be used directly.
        """
        self.io_sink.write(raw_content)

    def on_tracing_start(self, called: CallEnd):
        """
        Writes the diagram content leading to the first call (to the decorated function).
        This may be a no-operation.
        """
        raise NotImplementedError()

    def on_start_call(self, caller: CallEnd, called: CallEnd):
        """
        Writes the diagram content corresponding to a call between:
        - a caller: the block of code being executed
        - a called: a block of code being called
        These blocks of code are functions or methods.
        """
        raise NotImplementedError()

    def format_arg_value(self, arg: Any) -> str:
        """
        Formats a value being included in the diagram contents.
        Typically, a value returned by a function.
        Special care must be taken for the None value.
        """
        raise NotImplementedError()

    def on_error_propagation(self, error_called: CallEnd, error_caller: CallEnd, error: Error):
        """
        Writes the diagram contents corresponding to an error propagating
        between the called (which raised or propagated the error) and its caller
        """
        raise NotImplementedError()

    def on_return(self, *, called: CallEnd, caller: CallEnd, arg: Any):
        """
        Writes the diagram contents corresponding to a value returned
        by a called block of code to a caller block of code.
        """
        raise NotImplementedError()

    def on_tracing_end(self, called: CallEnd, arg: Any):
        """
        Special case of on_return when the last executed function returns a value.
        """
        raise NotImplementedError()

    def on_unhandled_error_end(self, called: CallEnd, error: Error):
        """
        Special case of on_tracing_end when an error was not handled by an except block.
        """
        raise NotImplementedError()

    def on_footer(self):
        """
        Writes the footer of the sequence diagram file.
        This method is supposed to be called only once during an export process.
        """
        raise NotImplementedError()

    @staticmethod
    def _template_dynamic_tags(export_file_path_template: str, moment: datetime) -> str:
        return Template(export_file_path_template).safe_substitute(
            datetime_millis=moment.strftime('%Y-%m-%d_%H.%M.%S_%f')[:-3]
        )

    @classmethod
    @contextmanager
    def export_manager_factory(exporter_class: 'Exporter', export_file_path: str) -> Iterator['Exporter']:
        """
        This factory function is meant to be used as a context manager to provide
        an exporter instance usable during the lifetime of the export context.

        The lifetime of the exporter instance is bound to the one of the context, which often
        involves a file handler resource (the export file being written) that is closed at
        the end of the context.
        """

        # hydrates datetime markers in the file name template
        export_file_path = exporter_class._template_dynamic_tags(export_file_path, datetime.utcnow())

        # creates the directories leading to the file
        Path(export_file_path).parent.mkdir(parents=True, exist_ok=True)

        # opens the contents file in write mode and yields the exporter that can write into it
        with open(export_file_path, 'w', encoding='utf8') as diagram_file:
            exporter = exporter_class(diagram_file)
            yield exporter


class Context(NamedTuple):
    """
    Stores information about the tracing process.
    It is used to pass information between the tracer and the diagram exporter.
    """

    exporter_class: Type[Exporter]
    export_file_path: str
    start_module: str
    start_function_name: str
    call_filter: CallFilter



"""
Domain models completing the execution domain models while exporting diagrams.

NamedTuples are used for immutability sake and for their light weight.
"""

from typing import Dict, Iterable, NamedTuple, Tuple, Union


class Function(NamedTuple):
    """
    Models a function (calling or being called):
    - it has a name
    - it is hosted in a module, modeled by its path as a tuple of strings
    """

    name: str
    module_path: Tuple[str]

    @property
    def fqn(self) -> str:
        return '.'.join((*self.module_path, self.name))


class Call(NamedTuple):
    """
    Flags the rank in the execution sequence when the execution leaves a block of code,
    when calling another block of code (another function).
    """

    rank: int


class Return(NamedTuple):
    """
    Flags the rank in the execution sequence when the execution exits a block of code normally,
    in an explicit or implicit return clause.
    """

    rank: int


class Raised(NamedTuple):
    """
    Flags the rank in the execution sequence when the execution exits a block of code abnormally,
    when an error is raised and not handled in the current block of code.
    """

    rank: int
    error: str


class Interactions(NamedTuple):
    calls: Iterable[Call]
    responses: Iterable[Union[Return, Raised]]


class Module(NamedTuple):
    name: str
    sub_modules: Dict[str, 'Module']
    functions: Dict[str, Function]



"""
Module dedicated to the export of component diagrams in the PlanUML syntax.

Bibliography:
- https://plantuml.com/en/component-diagram: syntax of component diagrams
The domain modeling involves NamedTuples because they are used as dictionary keys (thus need to be hashable and immutable).
"""

from collections import defaultdict, deque
from io import TextIOBase
from itertools import count
from string import Formatter
from typing import Any, Dict, Iterable, Tuple, Union

from pydoctrace.domain.diagram import Call, Function, Interactions, Module, Raised, Return
from pydoctrace.domain.execution import CallEnd, Error
from pydoctrace.exporters import Exporter
from pydoctrace.exporters.formatters import escape_dunder_with_tilde, formatter_factory
from pydoctrace.exporters.plantuml import FOOTER_TPL

PLANTUML_COMPONENT_FORMATTER: Formatter = formatter_factory('PlantUMLComponentFormatter', escape_dunder_with_tilde)

HEADER_TPL = r"""@startuml {diagram_name}
skinparam BoxPadding 10
skinparam componentStyle rectangle

"""

PACKAGE_OPEN_TPL = r"""{indentation}package {package_name:dunder} {{
"""

STYLED_PACKAGE_OPEN_TPL = r"""{indentation}{package_type} {package_name:dunder}{package_styling} {{
"""

PACKAGE_CLOSE_TPL = r"""{indentation}}}
"""

COMPONENT_TPL = r"""{indentation}[{function.fqn:dunder}] as "{function.name:dunder}"{stereotype}
"""

UNHANDLED_ERROR_LABEL_TPL = r"""{indentation}label {fq_error_class_name} as " "
"""

INTERACTION_CALL_TPL = r"""[{caller_function.fqn:dunder}] {arrow} [{called_function.fqn:dunder}]{arrow_label}
"""

UNHANDLED_ERROR_INTERACTION_TPL = r"""[{traced_function.fqn:dunder}] .up.> {fq_error_class_name} #line:darkred;text:darkred : {error_class_name:dunder}
"""

INDENT = '  '


class ModuleStructureVisitor:
    """
    Recursively produces the PlantUML syntax of the components structure using a visitor pattern on the root module.
    """

    fmt: Formatter = PLANTUML_COMPONENT_FORMATTER

    def __init__(self, traced_function: Function, unhandled_error_class_name: str):
        self.traced_function = traced_function
        self.unhandled_error_class_name = unhandled_error_class_name

    def visit_module(self, module: Module, parent_module_path: Tuple[str], indentation_level: int) -> Iterable[str]:
        """
        Yields the PlantUML code dedicated to package or modules hierarchy.
        """
        has_functions = len(module.functions) > 0

        # groups the module name with the parent ones if the module contains no function and only one sub-module
        if not has_functions and len(module.sub_modules) == 1:
            sub_module = list(module.sub_modules.values())[0]

            # skips the wrapping module if it has no name (root module)
            sub_module_parent_path = parent_module_path if module.name is None else parent_module_path + (module.name,)

            yield from self.visit_module(sub_module, sub_module_parent_path, indentation_level)

        # writes the parent packages if any,
        # then revisits the current module (without parents, with an increased indentation level)
        elif len(parent_module_path) > 0:
            indentation = indentation_level * INDENT

            yield self.fmt.format(PACKAGE_OPEN_TPL, indentation=indentation, package_name='.'.join(parent_module_path))

            yield from self.visit_module(module, (), indentation_level + 1)

            yield self.fmt.format(PACKAGE_CLOSE_TPL, indentation=indentation)
        # writes the module with its functions and sub-modules, or the package with sub-modules
        else:
            indentation = indentation_level * INDENT

            # opens the component
            # - special case when the root module contains more than one module: hide its contours (PlantUML rectangle)
            package_styling = ''
            if indentation_level == 0:
                package_type = 'rectangle'
                package_styling = ' #line:transparent;text:transparent'
            # - the component is a module if it has functions -> use a PlantUML frame representation
            elif has_functions:
                package_type = 'frame'
            # - the component is a package if it has no function -> use a PlantUML package representation
            else:
                package_type = 'package'

            yield self.fmt.format(
                STYLED_PACKAGE_OPEN_TPL,
                indentation=indentation,
                package_type=package_type,
                package_name=module.name,
                package_styling=package_styling,
            )

            # declares the functions as components
            if has_functions:
                yield from self.visit_functions(module.functions.values(), indentation_level + 1)

            # visits the sub-modules
            if len(module.sub_modules) > 0:
                sub_indentation_level = indentation_level + 1
                for sub_module in module.sub_modules.values():
                    yield from self.visit_module(sub_module, (), sub_indentation_level)

            # closes the component
            yield self.fmt.format(PACKAGE_CLOSE_TPL, indentation=indentation)

    def visit_functions(self, functions: Iterable[Function], indentation_level: int) -> Iterable[str]:
        """
        Yields the PlantUML code dedicated to the functions of a module.

        All functions are handled in the same visit because:
        - they share the same indentation, being the functions of the same module
        - the recursive visit stops here, being the branch leaves of the module hierarchy
        """
        indentation = indentation_level * INDENT

        for function in functions:
            is_traced_function = function == self.traced_function

            # adds the placeholder indicating that an unhandled error bubbles up from the traced execution
            if is_traced_function and self.unhandled_error_class_name is not None:
                yield self.fmt.format(
                    UNHANDLED_ERROR_LABEL_TPL,
                    indentation=indentation,
                    fq_error_class_name='.'.join([*self.traced_function.module_path, self.unhandled_error_class_name]),
                )

            yield self.fmt.format(
                COMPONENT_TPL,
                indentation=indentation,
                function=function,
                stereotype=' << @trace_to_component_puml >>' if is_traced_function else '',
            )


class PlantUMLComponentExporter(Exporter):
    """
    Exports the component diagram in the PlantUML format.

    The PlantUML syntax of the component diagram expects the components structure to be declared
    before the arrows representing the calls. Therefore, this exporter must store all the traced calls
    before exporting the diagram in order to build the components structure from the traced calls,
    then to export the arrows representing the calls.
    """

    fmt: Formatter = PLANTUML_COMPONENT_FORMATTER

    def __init__(self, io_sink: TextIOBase):
        super().__init__(io_sink)
        self.interactions_by_call: Dict[Tuple[Function, Function], Interactions] = defaultdict(
            lambda: Interactions(deque(), deque())
        )
        self.interaction_rank_iter = count(1, step=1)
        self.traced_function: Function = None
        self.functions: Dict[Tuple[str], Function] = {}
        self.unhandled_error_class_name: str = None

    def next_interaction_rank(self) -> int:
        return next(self.interaction_rank_iter)

    def function_from_call(self, caller: CallEnd) -> Function:
        """
        Retrieves the Function from the cache, or creates it from the given Call and caches it.
        """
        function_key = caller.fq_module_tuple + (caller.function_name,)
        function = self.functions.get(function_key)
        if function is None:
            function = Function(caller.function_name, caller.fq_module_tuple)
            self.functions[function_key] = function

        return function

    def add_call_interaction(self, caller: Function, called: Function):
        (self.interactions_by_call[caller, called]).calls.append(Call(self.next_interaction_rank()))

    def add_return_interaction(self, called: Function, caller: Function):
        (self.interactions_by_call[caller, called]).responses.append(Return(self.next_interaction_rank()))

    def add_raised_interaction(self, called: Function, caller: Function, error_class_name: str):
        (self.interactions_by_call[caller, called]).responses.append(
            Raised(self.next_interaction_rank(), error_class_name)
        )

    def on_header(self, start_module: str, start_func_name: str):
        diagram_name = f'{start_module}.{start_func_name}-component'
        self.io_sink.write(self.fmt.format(HEADER_TPL, diagram_name=diagram_name))

    def on_tracing_start(self, called: CallEnd):
        self.traced_function = self.function_from_call(called)

    def on_start_call(self, caller: CallEnd, called: CallEnd):
        self.add_call_interaction(self.function_from_call(caller), self.function_from_call(called))

    def on_error_propagation(self, error_called: CallEnd, error_caller: CallEnd, error: Error):
        self.add_raised_interaction(
            self.function_from_call(error_called), self.function_from_call(error_caller), error.class_name
        )

    def on_return(self, called: CallEnd, caller: CallEnd, **kwargs):
        self.add_return_interaction(self.function_from_call(called), self.function_from_call(caller))

    def on_tracing_end(self, called: CallEnd, arg: Any):
        pass

    def on_unhandled_error_end(self, called: CallEnd, error: Error):
        self.unhandled_error_class_name = error.class_name

    def on_footer(self):
        """
        At this stage, the exporter has all the information it needs to produce the contents of the diagram file
        """
        # builds components structure
        root_module = self.build_components_structure(self.functions.values())

        # writes components structure
        self.write_components_structure(root_module)

        # writes the arrow representing the unhandled error bubbling out of the tracing context
        self.write_unhandled_error_exit_interaction()

        # writes interactions
        self.write_components_interactions()

        # writes the file footer
        self.io_sink.write(FOOTER_TPL)

    def build_components_structure(self, functions: Iterable[Function]) -> Module:
        """
        Creates the modules hierarchy with their functions from the calls
        traced during the execution.
        """
        root_module = Module(None, {}, {})

        for function in functions:
            # creates or finds the modules hierarchy holding the function
            parent_module = root_module
            for module_name in function.module_path:
                sub_module = parent_module.sub_modules.get(module_name)
                if sub_module is None:
                    sub_module = Module(module_name, {}, {})
                    parent_module.sub_modules[module_name] = sub_module
                parent_module = sub_module

            # adds the function in the module
            sub_module.functions[function.name] = function

        return root_module

    def write_components_structure(self, root_module: Module):
        """
        Writes the 1st part of the PlantUML component diagram describing the structure of
        packages and modules, with the functions in them declared as UML components.
        """
        for component_line in ModuleStructureVisitor(
            self.traced_function, self.unhandled_error_class_name
        ).visit_module(root_module, (), 0):
            self.io_sink.write(component_line)

    def write_unhandled_error_exit_interaction(self):
        """
        Writes the interaction arrow between the traced function and the label
        representing the error that bubbles out of the traced function
        """
        if self.unhandled_error_class_name is not None:
            self.io_sink.write(
                self.fmt.format(
                    UNHANDLED_ERROR_INTERACTION_TPL,
                    traced_function=self.traced_function,
                    fq_error_class_name='.'.join([*self.traced_function.module_path, self.unhandled_error_class_name]),
                    error_class_name=self.unhandled_error_class_name,
                )
            )

    def build_arrow_label_ranks(self, ranks: Iterable[Union[int, str]]) -> str:
        if len(ranks) > 7:
            return f'{self.build_arrow_label_ranks(ranks[:3])} ... {self.build_arrow_label_ranks(ranks[-3:])}'
        else:
            return ', '.join(str(rank) for rank in ranks)

    def write_components_interactions(self):
        """
        Writes the 2nd part of the PlantUML component diagram describing the calls between
        the functions as arrows between the components.
        """
        for (caller_function, called_function), (calls, responses) in self.interactions_by_call.items():
            is_recursive_call = caller_function == called_function
            are_in_same_module = caller_function.module_path == called_function.module_path

            # call arrows between functions have a different direction whether they are in the same module or not
            call_arrow = '->' if is_recursive_call or not are_in_same_module else '-->'

            call_label = f' : {self.build_arrow_label_ranks([call.rank for call in calls])}'
            self.io_sink.write(
                self.fmt.format(
                    INTERACTION_CALL_TPL,
                    caller_function=caller_function,
                    arrow=call_arrow,
                    called_function=called_function,
                    arrow_label=call_label,
                )
            )

            # returns are not drawn for recursive calls to improve readability
            if not is_recursive_call:
                returns = [response for response in responses if isinstance(response, Return)]
                if returns:
                    returns_arrow = '<..' if are_in_same_module else '<.'
                    returns_label = f' : {self.build_arrow_label_ranks([exit_return.rank for exit_return in returns])}'
                    self.io_sink.write(
                        self.fmt.format(
                            INTERACTION_CALL_TPL,
                            caller_function=caller_function,
                            arrow=returns_arrow,
                            called_function=called_function,
                            arrow_label=returns_label,
                        )
                    )

            raiseds = [response for response in responses if isinstance(response, Raised)]
            if raiseds:
                raiseds_arrow = f"{'<..' if are_in_same_module else '<.'}[thickness=2]"
                raiseds_label = f" #line:darkred;text:darkred : {self.build_arrow_label_ranks([f'{exit_raised.rank}:{exit_raised.error}' for exit_raised in raiseds])}"

                self.io_sink.write(
                    self.fmt.format(
                        INTERACTION_CALL_TPL,
                        caller_function=caller_function,
                        arrow=raiseds_arrow,
                        called_function=called_function,
                        arrow_label=raiseds_label,
                    )
                )
